import sys
import os
import json
import dill as pickle

from colored import fore, back, style
import colorama

from .game import Game
from .player import Player
from .maps import Map

clear=lambda: os.system('cls' if os.name == 'nt' else 'clear')

class Launcher:
    saveFilePath = os.path.join(os.path.abspath(os.path.dirname(__file__)), "save")
    saveListFilePath = os.path.join(saveFilePath, "saveList.json")
    game = None

    def checkSavePath(self):
        if not os.path.exists(self.saveFilePath):
            os.makedirs(self.saveFilePath)

    def loadSaveFiles(self):
        try:
            with open(self.saveListFilePath) as json_file:  
                saveList = json.load(json_file)
        except:
            saveList = {}
        
        out = []
        for (key, value) in saveList.items():
            out.append({
                "saveId": key,
                "name": value
            })

        return out

    def deleteSave(self, saveId):
        try:
            with open(self.saveListFilePath, "r") as json_file:
                saveList = json.load(json_file)
        except:
            return
        
        try:
            del saveList[saveId]
            with open(self.saveListFilePath, "w") as json_file:
                json.dump(saveList, json_file)
            os.remove(os.path.join(self.saveFilePath, saveId))
        except KeyError:
            pass

    def loadSave(self, load):
        loadFile = open(os.path.join(self.saveFilePath, load["saveId"]), "r")  
        load = json.load(loadFile)
        return load

    def printSaveList(self, saveList):
        for i, save in enumerate(saveList):
            print(f"{i + 1}) {save['name']}")

    def startNewGame(self):
        self.game = Game()
        print("Choose a name:")
        name = input()
        self.game.player = Player(name)
        self.game.map = Map()
        print(f"<I>ronman Mode? {style.DIM}<Enter> for no{style.RESET}")
        ironman = input()
        if len(ironman) > 0 and ironman[0].upper() == "I":
            self.game.ironman = True
        self.runGame()

    def runGame(self):
        while not self.game.playerQuit and not self.game.restart:
            self.game.nextTurn()

    def startGame(self):
        clear()
        gameStarted = False

        print(f"{style.BOLD}Welcome To The Dungeon of 1000 Rooms!\n")
        print(f"<L>oad, <N>ew{style.RESET}")
        
        while not gameStarted:
            action = input()
            if len(action) > 0:
                action = action[0].upper()
            if action == "L":            
                saves = self.loadSaveFiles()
                if saves and len(saves) > 0:
                    clear()
                    self.printSaveList(saves)
                    saveFile = None
                    while not saveFile:
                        sys.stdout.write(f"\n{style.BOLD}Choose a save file: ")
                        saveChoice = input()
                        try:
                            saveIndex = int(saveChoice) - 1
                            saveFile = saves[saveIndex]
                            load = self.loadSave(saveFile)
                            self.game = Game()
                            self.game.player = Player(load["player"]["name"], load["player"])
                            self.game.player.loadItems(load["items"])
                            for i in load["game"]:
                                setattr(self.game, i, load["game"][i])
                            self.game.map = Map(load["map"]["numFloors"], load["map"]["width"], load["map"])

                            if self.game.ironman:
                                self.deleteSave(self.game.saveId)
                                self.game.saveId = ""

                            gameStarted = True
                            self.runGame()
                        except IndexError:
                            print("Invalid item number")
                        except ValueError:
                            print("Please enter an integer")
                else:
                    print ("No save file found, let's start a new game instead!")
                    gameStarted = True
                    self.startNewGame()
            elif action == "N":
                gameStarted = True
                self.startNewGame()

def main(args=None):
    if args is None:
        args = sys.argv[1:]

    launcher = Launcher()
    while not launcher.game or not launcher.game.playerQuit:
        launcher.startGame()
    if launcher.game.playerQuit:
        clear()
        launcher.game.player.printHistory(launcher.game.turn)
        print()

if __name__ == "__main__":
    colorama.init()
    main()
    colorama.deinit()


class Utils:
    @staticmethod
    def colorPadding(string):
        count = 0
        colorSequence = False
        for a in string: 
            if a == '\x1b':
                colorSequence = True 
                count += 1
            elif colorSequence and a != 'm':
                count += 1
            elif colorSequence and a == 'm':
                count += 1
                colorSequence = False
        return count

    @staticmethod
    def printStats(data, cellWidth = 25):
        for row in data:
            out = ""
            for field in row:
                value = row[field]
                out += (field + ": " + value).ljust(cellWidth + Utils.colorPadding(value))
            print(out)

    @staticmethod
    def printTable(header, data, cellWidth = 20):
        headerRow = ""
        if type(cellWidth) is int:
            cellWidth = [cellWidth] * len(header)

        for i, column in enumerate(header):
            headerRow += column.ljust(cellWidth[i] + Utils.colorPadding(column))
        print(headerRow)

        for row in data:
            out = ""
            try:
                out += row["_color"]
            except KeyError:
                pass
            dataFields = [key for key in list(row.keys()) if key != "_color"]
            for i, field in enumerate(dataFields):
                value = row[field]
                out += (value).ljust(cellWidth[i] + Utils.colorPadding(value))
            print(out)

    @staticmethod
    def flattenTuple(tuple):
        return "-".join(tuple)


import random

from colored import fore, back, style

from .monster_list import MonsterList
from .creature import Creature
from .utils import Utils

class Monster(Creature):
    distribution = [0,0,0,0,0,0,0,1,1,1,1,1,2,2,2,1,3,3,3,4,4,5,5,6,6,7,7]

    def __init__(self, dungeonLevel, data = None):
        genlevel = random.randint(max(1, dungeonLevel - 1), dungeonLevel + 1)
        self.charges = 0
        self.chargeRate = 1
        self.isBoss = False
        self.quotes = None
        self.seen = False
        self.known = False
        isBoss = data and data["id"] < 0
        if data and not isBoss:
            for k in data:
                setattr(self, k, data[k])
            try:
                test = self.displayName
            except AttributeError:
                self.displayName = self.name
        else:
            monsterLevel = data["floor"] + 1 if isBoss else genlevel
            info = self.getMonster(monsterLevel)
            Creature.__init__(self, info)
            self.level = monsterLevel
            self.ac += 10
            self.displayName = self.name

            if isBoss:
                # generate boss monster
                self.isBoss = True

                descriptor = random.choice(MonsterList.bossDescriptors[self.type])
                try:
                    descriptor = random.choice(MonsterList.bossDescriptors[self.subtype])
                except KeyError:
                    pass
                if descriptor[0]:
                    self.displayName = f"{descriptor[0]} {self.displayName}"
                if descriptor[1]:
                    self.displayName += f" {descriptor[1]}"

                self.quotes = MonsterList.bossQuotes[self.type]
                try:
                    self.quotes = MonsterList.bossQuotes[self.subtype]
                except KeyError:
                    pass

                diffFactor = max(dungeonLevel, data["floor"]) + 2
                self.chargeRate = 2
                self.level += diffFactor
                self.atk += diffFactor
                self.ac += diffFactor
                self.maxHp = self.hd * self.level
                self.hp = self.hd * self.level
            else:
                # improved monsters
                levelDiff = dungeonLevel - self.level
                if levelDiff > 0:
                    descriptor = MonsterList.descriptors[self.type][levelDiff // 2]
                    try:
                        descriptor = MonsterList.descriptors[self.subtype][levelDiff // 2]
                    except KeyError:
                        pass
                    self.displayName = f"{descriptor} {self.displayName}"
                    self.level += levelDiff
                    self.atk += 2 * levelDiff
                    self.ac += 2 * levelDiff    
                    self.hd += levelDiff    

                self.maxHp = self.hd * self.level
                self.hp = 0
                for x in range(self.level):
                    self.hp += random.randint(self.hd // 2, self.hd)
                                    
    def printStats(self, playerLore):
        nameColor = fore.DARK_ORANGE_3B if self.isBoss else fore.RED
        print(f"{nameColor}{style.BOLD}{self.displayName} ({str(self.level)}){style.RESET}")
        try:
            lore = playerLore[str(self.id)]
        except KeyError:     
            lore = { "resist": False, "vulnerability": False, "special": False}

        unknown = f"{style.DIM}???{style.RESET}"
        resist = self.resist if self.resist else "---"
        vulnerability = self.vulnerability if self.vulnerability else "---"
        special = self.special if self.special else "extra attack"

        stats = [
            {
                "HP": f"{self.hp}",
                "ATK": f"{self.atk} {Creature.calculateDam(self)}",
                "AC": f"{self.ac}"
            },
            {
                "Resists": resist if lore["resist"] else unknown,
                "Vulnerability": vulnerability if lore["vulnerability"] else unknown,
                "Special Attack": special if lore["special"] else unknown
            }
        ]
        Utils.printStats(stats)

    def getAtkVerb(self):
        verbs = MonsterList.atkVerbs[self.atk_type]
        try:
            verbs = MonsterList.atkVerbs[self.type][self.atk_type]
        except KeyError:
            pass
        try:
            verbs = MonsterList.atkVerbs[self.subtype][self.atk_type]
        except KeyError:
            pass
        return random.choice(verbs)

    @staticmethod
    def getMonster(level):
        randomLevel = 0
        while randomLevel <= 0:
            randomLevel = level - random.choice(Monster.distribution)
        monsters = [monster for monster in MonsterList.monsters if monster["level"] == randomLevel] 
        info = random.choice(monsters)
        return info




from colored import fore, back, style

from .utils import Utils

class Creature:
    def __init__(self, info):
        for i in info:
            setattr(self, i, info[i])

        self.calculateDam()

    def calculateDam(self):
        return f"{style.DIM}({self.atk + 2}-{self.atk + (self.level * 2)}){style.RESET}"

    def damage(self, value, type):
        damage = value
        try:
            if type in self.resist:
                damage = value // 2
        except KeyError:
            pass
        except AttributeError:
            pass
        try:
            if self.vulnerability == type:
                damage = value * 2
        except KeyError:
            pass
        except AttributeError:
            pass
        self.hp -= damage
        return damage
        
    def heal(self, value = 0):
        if value == 0:
            self.hp = self.maxHp
        else :
            self.hp = min(self.hp + value, self.maxHp)